@@ -6,10 +6,15 @@ Pyro - Python Remote Objects.  Copyright by Irmen de Jong (irmen@razorvine.net).
 
 import sys
 import zlib
+import uuid
 import logging
 import linecache
 import traceback
 import inspect
+import struct
+import datetime
+import decimal
+import numbers
 from Pyro4 import errors
 from Pyro4.configuration import config
 
@@ -139,6 +144,7 @@ else:
     for name, t in vars(builtins).items():
         if type(t) is type and issubclass(t, BaseException):
             all_exceptions[name] = t
+    buffer = bytearray
 for name, t in vars(errors).items():
     if type(t) is type and issubclass(t, errors.PyroError):
         all_exceptions[name] = t
@@ -158,6 +164,8 @@ class SerializerBase(object):
     def deserializeData(self, data, compressed=False):
         """Deserializes the given data (bytes). Set compressed to True to decompress the data first."""
         if compressed:
+            if sys.version_info < (3, 0):
+                data = self._convertToBytes(data)
             data = zlib.decompress(data)
         return self.loads(data)
 
@@ -171,6 +179,8 @@ class SerializerBase(object):
         """Deserializes the given call data back to (object, method, vargs, kwargs) tuple.
         Set compressed to True to decompress the data first."""
         if compressed:
+            if sys.version_info < (3, 0):
+                data = self._convertToBytes(data)
             data = zlib.decompress(data)
         return self.loadsCall(data)
 
@@ -186,6 +196,15 @@ class SerializerBase(object):
     def dumpsCall(self, obj, method, vargs, kwargs):
         raise NotImplementedError("implement in subclass")
 
+    def _convertToBytes(self, data):
+        t = type(data)
+        if t is not bytes:
+            if t in (bytearray, buffer):
+                return bytes(data)
+            if t is memoryview:
+                return data.tobytes()
+        return data
+
     def __compressdata(self, data, compress):
         if not compress or len(data) < 200:
             return data, False  # don't waste time compressing small messages
@@ -301,6 +320,7 @@ class SerializerBase(object):
         Only a fixed set of classes are recognized.
         Not used for the pickle serializer.
         """
+        from Pyro4 import core, futures  # XXX circular
         classname = data.get("__class__", "<unknown>")
         if isinstance(classname, bytes):
             classname = classname.decode("utf-8")
@@ -311,7 +331,6 @@ class SerializerBase(object):
             raise errors.SecurityError("refused to deserialize types with double underscores in their name: " + classname)
         # for performance, the constructors below are hardcoded here instead of added on a per-class basis to the dict-to-class registry
         if classname.startswith("Pyro4.core."):
-            from Pyro4 import core  # XXX circular
             if classname == "Pyro4.core.URI":
                 uri = core.URI.__new__(core.URI)
                 uri.__setstate_from_dict__(data["state"])
@@ -325,23 +344,26 @@ class SerializerBase(object):
                 daemon.__setstate_from_dict__(data["state"])
                 return daemon
         elif classname.startswith("Pyro4.util."):
-            if classname == "Pyro4.util.PickleSerializer":
+            if classname == "Pyro4.util.SerpentSerializer":
+                return SerpentSerializer()
+            elif classname == "Pyro4.util.PickleSerializer":
                 return PickleSerializer()
-            elif classname == "Pyro4.util.DillSerializer":
-                return DillSerializer()
             elif classname == "Pyro4.util.MarshalSerializer":
                 return MarshalSerializer()
             elif classname == "Pyro4.util.JsonSerializer":
                 return JsonSerializer()
-            elif classname == "Pyro4.util.SerpentSerializer":
-                return SerpentSerializer()
+            elif classname == "Pyro4.util.MsgpackSerializer":
+                return MsgpackSerializer()
+            elif classname == "Pyro4.util.DillSerializer":
+                return DillSerializer()
         elif classname.startswith("Pyro4.errors."):
             errortype = getattr(errors, classname.split('.', 2)[2])
             if issubclass(errortype, errors.PyroError):
                 return SerializerBase.make_exception(errortype, data)
         elif classname == "Pyro4.futures._ExceptionWrapper":
-            from Pyro4 import futures  # XXX circular
-            ex = SerializerBase.dict_to_class(data["exception"])
+            ex = data["exception"]
+            if isinstance(ex, dict) and "__class__" in ex:
+                ex = SerializerBase.dict_to_class(ex)
             return futures._ExceptionWrapper(ex)
         elif data.get("__exception__", False):
             if classname in all_exceptions:
@@ -364,11 +386,6 @@ class SerializerBase(object):
                 exceptiontype = getattr(sqlite3, short_classname)
                 if issubclass(exceptiontype, BaseException):
                     return SerializerBase.make_exception(exceptiontype, data)
-
-        # try one of the serializer classes
-        for serializer in _serializers.values():
-            if classname == serializer.__class__.__name__:
-                return serializer
         log.warning("unsupported serialized class: " + classname)
         raise errors.SerializeError("unsupported serialized class: " + classname)
 
@@ -422,9 +439,11 @@ class PickleSerializer(SerializerBase):
         return pickle.dumps(data, config.PICKLE_PROTOCOL_VERSION)
 
     def loadsCall(self, data):
+        data = self._convertToBytes(data)
         return pickle.loads(data)
 
     def loads(self, data):
+        data = self._convertToBytes(data)
         return pickle.loads(data)
 
     @classmethod
@@ -481,22 +500,39 @@ class MarshalSerializer(SerializerBase):
         except (ValueError, TypeError):
             return marshal.dumps(self.class_to_dict(data))
 
-    def loadsCall(self, data):
-        obj, method, vargs, kwargs = marshal.loads(data)
-        vargs = self.recreate_classes(vargs)
-        kwargs = self.recreate_classes(kwargs)
-        return obj, method, vargs, kwargs
-
     if sys.platform == "cli":
+        def loadsCall(self, data):
+            if type(data) is not str:
+                # Ironpython's marshal expects str...
+                data = str(data)
+            obj, method, vargs, kwargs = marshal.loads(data)
+            vargs = self.recreate_classes(vargs)
+            kwargs = self.recreate_classes(kwargs)
+            return obj, method, vargs, kwargs
+
         def loads(self, data):
             if type(data) is not str:
                 # Ironpython's marshal expects str...
                 data = str(data)
             return self.recreate_classes(marshal.loads(data))
     else:
+        def loadsCall(self, data):
+            data = self._convertToBytes(data)
+            obj, method, vargs, kwargs = marshal.loads(data)
+            vargs = self.recreate_classes(vargs)
+            kwargs = self.recreate_classes(kwargs)
+            return obj, method, vargs, kwargs
+
         def loads(self, data):
+            data = self._convertToBytes(data)
             return self.recreate_classes(marshal.loads(data))
 
+    @classmethod
+    def class_to_dict(cls, obj):
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        return super(MarshalSerializer, cls).class_to_dict(obj)
+
     @classmethod
     def register_type_replacement(cls, object_type, replacement_function):
         pass  # marshal serializer doesn't support per-type hooks
@@ -555,14 +591,14 @@ class JsonSerializer(SerializerBase):
         return data.encode("utf-8")
 
     def loadsCall(self, data):
-        data = data.decode("utf-8")
+        data = self._convertToBytes(data).decode("utf-8")
         data = json.loads(data)
         vargs = self.recreate_classes(data["params"])
         kwargs = self.recreate_classes(data["kwargs"])
         return data["object"], data["method"], vargs, kwargs
 
     def loads(self, data):
-        data = data.decode("utf-8")
+        data = self._convertToBytes(data).decode("utf-8")
         return self.recreate_classes(json.loads(data))
 
     def default(self, obj):
@@ -571,6 +607,12 @@ class JsonSerializer(SerializerBase):
             obj = replacer(obj)
         if isinstance(obj, set):
             return tuple(obj)  # json module can't deal with sets so we make a tuple out of it
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        if isinstance(obj, (datetime.datetime, datetime.date)):
+            return obj.isoformat()
+        if isinstance(obj, decimal.Decimal):
+            return str(obj)
         return self.class_to_dict(obj)
 
     @classmethod
@@ -578,6 +620,73 @@ class JsonSerializer(SerializerBase):
         cls.__type_replacements[object_type] = replacement_function
 
 
+class MsgpackSerializer(SerializerBase):
+    """(de)serializer that wraps the msgpack serialization protocol."""
+    serializer_id = 6  # never change this
+
+    __type_replacements = {}
+
+    def dumpsCall(self, obj, method, vargs, kwargs):
+        return msgpack.packb((obj, method, vargs, kwargs), use_bin_type=True, default=self.default)
+
+    def dumps(self, data):
+        return msgpack.packb(data, use_bin_type=True, default=self.default)
+
+    def loadsCall(self, data):
+        data = self._convertToBytes(data)
+        obj, method, vargs, kwargs = msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook)
+        return obj, method, vargs, kwargs
+
+    def loads(self, data):
+        data = self._convertToBytes(data)
+        return msgpack.unpackb(data, encoding="utf-8", object_hook=self.object_hook, ext_hook=self.ext_hook)
+
+    def default(self, obj):
+        replacer = self.__type_replacements.get(type(obj), None)
+        if replacer:
+            obj = replacer(obj)
+        if isinstance(obj, set):
+            return tuple(obj)  # msgpack module can't deal with sets so we make a tuple out of it
+        if isinstance(obj, uuid.UUID):
+            return str(obj)
+        if isinstance(obj, bytearray):
+            return bytes(obj)
+        if isinstance(obj, complex):
+            return msgpack.ExtType(0x30, struct.pack("dd", obj.real, obj.imag))
+        if isinstance(obj, datetime.datetime):
+            if obj.tzinfo:
+                raise errors.SerializeError("msgpack cannot serialize datetime with timezone info")
+            return msgpack.ExtType(0x32, struct.pack("d", obj.timestamp()))
+        if isinstance(obj, datetime.date):
+            return msgpack.ExtType(0x33, struct.pack("l", obj.toordinal()))
+        if isinstance(obj, decimal.Decimal):
+            return str(obj)
+        if isinstance(obj, numbers.Number):
+            return msgpack.ExtType(0x31, str(obj).encode("ascii"))     # long
+        return self.class_to_dict(obj)
+
+    def object_hook(self, obj):
+        if "__class__" in obj:
+            return self.dict_to_class(obj)
+        return obj
+
+    def ext_hook(self, code, data):
+        if code == 0x30:
+            real, imag = struct.unpack("dd", data)
+            return complex(real, imag)
+        if code == 0x31:
+            return int(data)
+        if code == 0x32:
+            return datetime.datetime.fromtimestamp(struct.unpack("d", data)[0])
+        if code == 0x33:
+            return datetime.date.fromordinal(struct.unpack("l", data)[0])
+        raise errors.SerializeError("invalid ext code for msgpack: "+str(code))
+
+    @classmethod
+    def register_type_replacement(cls, object_type, replacement_function):
+        cls.__type_replacements[object_type] = replacement_function
+
+
 """The various serializers that are supported"""
 _serializers = {}
 _serializers_by_id = {}
@@ -634,14 +743,21 @@ try:
     else:
         ver = serpent.__version__
     ver = tuple(map(int, ver.split(".")))
-    if ver < (1, 16):
-        raise RuntimeError("requires serpent 1.17 or better")
+    if ver < (1, 19):
+        raise RuntimeError("requires serpent 1.19 or better")
     _ser = SerpentSerializer()
     _serializers["serpent"] = _ser
     _serializers_by_id[_ser.serializer_id] = _ser
 except ImportError:
     log.warning("serpent serializer is not available")
     pass
+try:
+    import msgpack
+    _ser = MsgpackSerializer()
+    _serializers["msgpack"] = _ser
+    _serializers_by_id[_ser.serializer_id] = _ser
+except ImportError:
+    pass
 del _ser
 
 
